#! /usr/bin/python
from __future__ import print_function

'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

"""
python package to formulate and solve resource-constrained scheduling problems
"""



from collections import OrderedDict as _DICT_TYPE
import types
import functools
import copy
import uuid

try: # allow Python 2 vs 3 compatibility
	_maketrans = str.maketrans
except AttributeError:
	from string import maketrans as _maketrans

def _isnumeric(var) :
	"""
	Test if var is numeric, only integers are allowed
	"""
	return type(var) is int

def _isiterable(var) :
	"""
	Test if var is iterable
	"""
	return ( type(var) is list ) or ( type(var) is set ) or ( type(var) is tuple )

def alt(*args) :
	"""
	Method to reduce the given elements with the or-operator
	e.g. alt(R1,R2,R3) = alt( R for R in [R1,R2,R3] ) = alt([R1,R2,R3]) = R1|R2|R3
	"""
	l = [ functools.reduce(lambda x,y: x|y, a) for a in args if _isiterable(a) or isinstance(a, types.GeneratorType)]
	l += [ a for a in args if not _isiterable(a) and not isinstance(a, types.GeneratorType)  ]
	return functools.reduce(lambda x,y: x|y, l)



class _SchedElement(object):

	def __init__(self,name='') :
		if type(name) is not str:
			raise Exception('ERROR: name %s is not a string'%str(name))
		if 'start' in name or 'end' in name:
			raise Exception('ERROR: avoid the substring "start" and "end" in any names, this will cause problems with solvers')
		trans = _maketrans("-+[] ->/","________")
		if name.translate(trans) != name:
			raise Exception('ERROR: name %s contains one of the following characters: -+[] ->/'%name)
		self.name = name

	def __str__(self) :
		return str(self.name)

	def __repr__(self) :
		return self.__str__()

	def __hash__(self) :
		return self.name.__hash__()



class _SchedElementAffine(_DICT_TYPE) :

	def __init__(self,unknown=None,element_class=_SchedElement,affine_operator='+') :
		_DICT_TYPE.__init__(self)
		self.element_class = element_class
		self.affine_operator = affine_operator
		if unknown == None :
			pass
		elif isinstance(unknown,self.element_class) :
			self[unknown] = 1
		elif _isnumeric(unknown) :
			self[1] = unknown
		elif isinstance(unknown,type(self)) :
			self.update(unknown)
		elif isinstance(unknown,list) :
			self.update(_DICT_TYPE(unknown))
		else :
			raise Exception('ERROR: cant init %s from %s'%(str(self),str(unknown)))

	def __add__(self,other) :
		if _isiterable(other) :
			return [ self + x for x in other ]
		if isinstance(other,type(self)) :
			new = type(self)(self)
			for key in other :
				if key in new :
					new[key] += other[key]
				else :
					new[key] = other[key]
			return new
		elif isinstance(other,self.element_class) or _isnumeric(other):
			return self + type(self)(other)
		raise Exception('ERROR: you cannot add %s to %s'%(str(other),str(self)))

	def __sub__(self,other) :
		if _isiterable(other) :
			return [ self - x for x in other ]
		if isinstance(other,type(self)) :
			new = type(self)(self)
			for key in other :
				if key in new :
					new[key] -= other[key]
				else :
					new[key] = -other[key]
			return new
		else :
			return self - type(self)(other)

	def __mul__(self,other) :
		if isinstance(other,type(self)) :
			new = type(self)(self)
			if 1 in other :
				for key in self :
					new[key] *= other[1]
			else :
				raise Exception('ERROR: can multipy %s only with integer but not with %s'%(str(self),str(other)))
			return new
		else :
			return self * type(self)(other)

	def __radd__(self,other) :
		return self + other

	def __str__(self) :
		def format_coeff(val) :
			if val != 1 :
				return str(val)+'x'
			return ''
		return self.affine_operator.join([ format_coeff(self[key])+str(key) for key in self ])

	def __repr__(self):
		return self.__str__() + '_' + str(id(self)) #add id because representation is otherwise not unique

	def __hash__(self) :
		return self.__repr__().__hash__()



class Scenario(_SchedElement):
	"""
	The base scenario class
	"""
	def __init__(self,name='scenario not named',horizon=None):
		_SchedElement.__init__(self,name)
		self.horizon = horizon
		self._tasks = _DICT_TYPE() #tasks
		self._resources = _DICT_TYPE() #resources
		self._constraints = list()

	def Task(self,name,length=1,group=None,reward=None,**kwargs) :
		"""
		Adds a new task to the scenario
		name : unique task name, must not contain special characters
		length : length of task, default is 1
		"""
		if name in self._tasks or name in self._resources:
			raise Exception('ERROR: resource or task with name %s already contained in scenario'%str(name))
		task = Task(name=name,
			length=length,
			group=group,
			reward=reward,
			**kwargs)
		self.add_task(task)
		return task

	def tasks(self,resource=None) :
		"""
		Returns all tasks in scenario
		"""
		if resource is None :
			return list(self._tasks.values())
		else :
			return list({ T for T in self.tasks() for RA in T.resources_req if resource in RA })

	def Resource(self,name,size=1) :
		"""
		Adds a new resource to the scenario
		name   : unique resource name, must not contain special characters
		size   : the size of the resource, if size > 1, then we get a cumulative resource that can process
		 		different tasks in parallel
		"""
		if name in self._tasks or name in self._resources:
			raise Exception('ERROR: resource or task with name %s already contained in scenario'%str(name))
		resource = Resource(name,size=size)
		self.add_resource(resource)
		return resource

	def resources(self,task=None) :
		"""
		Returns all resources in scenario
		"""
		if task is None :
			return list(self._resources.values())
		else :
			return list({ R for RA in task.resources_req for R in RA })

	def resources_req_tasks(self,min_size=2):
		"""
		Returns a mapping of resource requirements to tasks. This is helpful if resources requirements
		are jointly used by different tasks. Only resource requiremenets
		min_size : the minimum number of tasks in resources req, default is 2
		"""
		ra_to_tasks = dict()
		for T in self.tasks():
			for RA in T.resources_req:
				if len(RA) < min_size:
					continue
				if RA not in ra_to_tasks:
					ra_to_tasks[RA] = {T}
				else:
					ra_to_tasks[RA].add(T)
		return ra_to_tasks

	def solution(self) :
		"""
		Returns the last computed solution in tabular form with columns: task, resource, start, end
		"""
		solution = \
			[ (T,R,T.start_value,T.start_value+T.length)
			for T in self.tasks()
			if T.start_value != None and T.resources != None
			for R in T.resources
			]

		solution = sorted(solution, key = lambda x : (x[2],str(x[0]),str(x[1])) ) # sort according to start and name
		return solution

	def objective(self):
		"""
		Returns a representation of all objectives
		"""
		tasks_objective = [T*T['completion_time_cost'] for T in self.tasks() if 'completion_time_cost' in T ]
		if tasks_objective:
			return functools.reduce(lambda x,y:x+y,tasks_objective)
		return None

	def objective_value(self) :
		"""
		Returns the value of the objective
		"""
		return sum([ T['_completion_time_cost']*(T.start_value+T.length) for T in self.tasks()
		 			if '_completion_time_cost' in T])

	def use_makespan_objective(self) :
		"""
		Set the objective to the makespan of all included tasks
		"""
		self.clear_objective()
		if not self.tasks():
			return
		if 'MakeSpan' in self._tasks :
			self._constraints = [ C for C in self._constraints if self._tasks['MakeSpan'] not in C.tasks() ]
			del self._tasks['MakeSpan']
		tasks = self.tasks() # save tasks before adding makespan
		makespan = self.Task('MakeSpan')
		makespan += self.resources()[0] # add first resource, every task needs one
		for T in tasks :
			self += T < makespan
		self += makespan*1

	def use_flowtime_objective(self) :
		"""
		Sets the objective to a uniform flow-time objective
		"""
		self.clear_objective()
		if not self.tasks():
			return
		A = sum([ T*1 for T in self.tasks() ])
		del A[1] #remove 1 due to sum
		self += A

	def clear_solution(self):
		"""
		Remove the solution from the scenario
		"""
		for T in self.tasks() :
			T.start_value = None
			T.resources = None

	def clear_objective(self):
		"""
		Removes all objective annotations
		"""
		for T in self.tasks():
			T.completion_time_cost = None

	def constraints(self,constraint_class=None):
		if constraint_class is None:
			return self._constraints
		return [ C for C in self._constraints if isinstance(C,constraint_class) ]

	def precs_lax(self):
		return self.constraints(PrecedenceLax)

	def precs_tight(self):
		return self.constraints(PrecedenceTight)

	def precs_cond(self):
		return self.constraints(PrecedenceCond)

	def bounds_low(self) :
		return self.constraints(BoundLow)

	def bounds_up(self) :
		return self.constraints(BoundUp)

	def bounds_low_tight(self) :
		return self.constraints(BoundLowTight)

	def bounds_up_tight(self) :
		return self.constraints(BoundUpTight)

	def capacity_low(self):
		return self.constraints(CapacityLow)

	def capacity_up(self):
		return self.constraints(CapacityUp)

	def capacity_diff_up(self):
		return self.constraints(CapacityDiffUp)

	def capacity_diff_low(self):
		return self.constraints(CapacityDiffLow)

	def add_constraint(self,constraint):
		for task in constraint.tasks():
			if task not in self:
				raise Exception('ERROR: task %s is not contained in scenario %s'%(str(task.name),str(self.name)))
		for resource in constraint.resources():
			if resource not in self:
				raise Exception('ERROR: resource %s is not contained in scenario %s'%(str(resource.name),str(self.name)))
		if str(constraint) in [ str(C) for C in self._constraints ]:
			return self
		self._constraints.append(constraint)

	def remove_constraint(self,constraint):
		if str(constraint) not in [ str(C) for C in self._constraints ]:
			raise Exception('ERROR: constraint %s not contained in scenario %s'%(str(constraint),str(self.name)))
		self._constraints = [ C for C in self._constraints if str(C) != str(constraint) ]

	def add_task(self,task):
		if task.name in self._tasks and task is not self._tasks[task.name]:
			raise Exception('ERROR: task with name %s already contained in scenario %s' % (str(task.name),str(self.name)))
		elif task not in self.tasks():
			self._tasks[task.name] = task

	def remove_task(self,task):
		if task.name in self._tasks :
			del self._tasks[task.name]
		else :
			raise Exception('ERROR: task with name %s not contained in scenario %s' % (str(task.name),str(self.name)))
		self._constraints = [ C for C in self._constraints if task not in C.tasks() ]

	def add_task_affine(self,task_affine):
		for task in task_affine:
			if isinstance(task,Task):
				if task not in self:
					raise Exception('ERROR: task %s is not contained in scenario %s'%(str(task.name),str(self.name)))
			task.completion_time_cost = task_affine[task]

	def add_resource(self,resource):
		if resource.name in self._resources and resource is not self._resources[resource.name]:
			raise Exception('ERROR: resource with name %s already contained in scenario %s'%
						(str(resource.name),str(self.name)))
		elif resource not in self.resources():
			self._resources[resource.name] = resource

	def remove_resource(self,resource):
		if resource.name in self._resources :
			del self._resources[resource.name]
		else:
			raise Exception('ERROR: resource with name %s not contained in scenario %s'%
						(str(resource.name),str(self.name)))
		self._constraints = [ C for C in self._constraints if resource not in C.resources() ]

	def __iadd__(self,other) :
		if _isiterable(other) :
			for x in other :
				self += x
			return self
		elif isinstance(other,_Constraint) :
			self.add_constraint(other)
			return self
		elif isinstance(other,Task) :
			self.add_task(other)
			return self
		elif isinstance(other,_TaskAffine) :
			self.add_task_affine(other)
			return self
		elif isinstance(other,Resource) :
			self.add_resource(self,other)
			return self
		raise Exception('ERROR: cant add %s to scenario %s'%(str(other),str(self.name)))

	def __isub__(self,other) :
		if _isiterable(other):
			for x in other:
				self -= x
		elif isinstance(other,Task) :
			self.remove_task(other)
		elif isinstance(other,Resource):
			self.remove_resource(other)
		elif isinstance(other,_Constraint):
			self.remove_constraint(other)
		else:
			raise Exception('ERROR: cant subtract %s to scenario %s'%(str(other),str(self.name)))
		return self

	def __contains__(self, item):
		if isinstance(item,Task):
			return item in self._tasks.values()
		elif isinstance(item,Resource):
			return item in self._resources.values()
		else:
			raise Exception('ERROR: %s cannot be checked for containment in scenario %s'%(str(item),str(self.name)))
		return self

	def __getitem__(self, item):
		if item not in self._tasks and item not in self._resources:
			raise Exception('ERROR: task or resource with name %s is not contained in scenario %s'%
						(str(item),str(self.name)))
		if item in self._tasks:
			return self._tasks[item]
		return self._resources[item]

	def __setitem__(self,key,item): # getitem does not work without having also setitem?
		return self

	def check(self):
		"""
		Do basic checks on scenario
		"""
		# check if each task has a resource
		for T in self.tasks():
			if not T.resources_req:
				raise Exception('ERROR: task %s has no resource requirement'%str(T))

	def __str__(self) :
		s = '###############################################\n'
		s += '\n'
		s += 'SCENARIO: '+self.name
		if self.horizon is not None:
			s += ' / horizon: %i\n\n'%self.horizon
		else:
			s += ' / no horizon set\n\n'

		s += 'OBJECTIVE: '+str(self.objective())+'\n\n'

		s += 'RESOURCES:\n'
		for R in self.resources() :
			s += str(R.name)+'\n'
		s += '\n'

		s += 'TASKS:\n'
		for T in self.tasks() :
			s += '%s : %s\n'%(str(T.name),','.join([ str(RA) for RA in T.resources_req ]))
		s += '\n'

		s += 'JOINT RESOURCES:\n'
		ra_to_tasks = self.resources_req_tasks()
		for RA in ra_to_tasks:
			if len(RA) < 2:
				continue
			s += '%s : %s\n'%(str(RA),','.join([ str(T) for T in ra_to_tasks[RA] ]))
		s += '\n'

		def print_constraint(title,constraints):
			s = ''
			if constraints:
				s += '%s:\n'%title
				s += '\n'.join([ C.__repr__() for C in constraints ]) + '\n'
				s += '\n'
			return s
		s += print_constraint('LAX PRECEDENCES',self.precs_lax())
		s += print_constraint('TIGHT PRECEDENCES',self.precs_tight())
		s += print_constraint('COND PRECEDENCES',self.precs_cond())
		s += print_constraint('LOWER BOUNDS',self.bounds_low())
		s += print_constraint('UPPER BOUNDS',self.bounds_up())
		s += print_constraint('TIGHT LOWER BOUNDS',self.bounds_low_tight())
		s += print_constraint('TIGHT UPPER BOUNDS',self.bounds_up_tight())
		s += print_constraint('CAPACITY LOWER BOUNDS',self.capacity_low())
		s += print_constraint('CAPACITY UPPER BOUNDS',self.capacity_up())
		s += print_constraint('CAPACITY DIFF LOWER BOUNDS',self.capacity_diff_low())
		s += print_constraint('CAPACITY DIFF UPPER BOUNDS',self.capacity_diff_up())
		s += '###############################################'
		return s



class Task(_SchedElement) :
	"""
	A task to be processed by at least one resource
	"""
	def __init__(self,name,length=1,group=None,reward=None,completion_time_cost=None,**kwargs) :
		_SchedElement.__init__(self,name)
		if not _isnumeric(length):
			raise Exception('ERROR: task length must be an integer')
		# base parameters
		self.length = length # length of task
		self.group = group # group exchangeable tasks

		# additional parameters
		self.start_value = None # should be filled by solver
		self.resources = None # should be filled by solver
		self.resources_req = list() # required resources
		self.completion_time_cost = completion_time_cost # cost on the final completion time
		self.reward = reward # in case not None, then the task is optional and substracts the reward from the cost

		for key in kwargs:
			self.__setattr__(key,kwargs[key])

	def __len__(self) :
		return self.length

	def __lt__(self,other) :
		return _TaskAffine(self) < other

	def __gt__(self,other) :
		return _TaskAffine(self) > other

	def __le__(self,other) :
		return _TaskAffine(self) <= other

	def __ge__(self,other) :
		return _TaskAffine(self) >= other

	def __ne__(self,other) :
		return _TaskAffine(self) != other

	def __lshift__(self,other) :
		return _TaskAffine(self) << other

	def __rshift__(self,other) :
		return _TaskAffine(self) >> other

	def __add__(self,other) :
		return _TaskAffine(self) + other

	def __sub__(self,other) :
		return _TaskAffine(self) - other

	def __mul__(self,other) :
		return _TaskAffine(self) * other

	def __radd__(self,other) :
		return _TaskAffine(self) + other

	def add_resources_req(self,RA):
		self.resources_req.append(RA)

	def remove_resources_req(self,RA):
		self.resources_req = [ RA_ for RA_ in self.resources_req if str(RA_) != str(RA) ]
		return self

	def __iadd__(self,other):
		if _isiterable(other):
			for x in other:
				self += x
			return self
		elif isinstance(other,Resource):
			other = _ResourceAffine(other) #transform into _ResourceAffine
			self.add_resources_req(other)
			return self
		elif isinstance(other,_ResourceAffine):
			self.add_resources_req(other)
			return self
		raise Exception('ERROR: cant add %s to task %s'%(str(other),str(self)))

	def __isub__(self,other):
		if _isiterable(other):
			for x in other:
				self += x
			return self
		elif isinstance(other,Resource):
			other = _ResourceAffine(other) #transform into _ResourceAffine
			self.remove_resources_req(other)
			return self
		elif isinstance(other,_ResourceAffine):
			self.remove_resources_req(other)
			return self
		raise Exception('ERROR: cant subtract %s from task %s'%(str(other),str(self)))

	def __setitem__(self, key, value):
		setattr(self,str(key),value)

	def __getitem__(self, key):
		return getattr(self,str(key))

	def __contains__(self,key):
		if not hasattr(self,key):
			return False
		if getattr(self,key) is None:
			return False
		return True



class _TaskAffine(_SchedElementAffine) :

	def __init__(self,unknown=None) :
		_SchedElementAffine.__init__(self,unknown=unknown,element_class=Task)

	def _get_prec(self,TA,comp_operator) :
		pos_tasks = [ T for T in TA if isinstance(T,Task) and TA[T] >= 0 ]
		neg_tasks = [ T for T in TA if isinstance(T,Task) and TA[T] < 0 ]
		if len(neg_tasks) > 1 or len(pos_tasks) > 1 :
			raise Exception('ERROR: can only deal with simple precedences of \
									the form T1 + 3 < T2 or T1 < 3 and not %s'%str(TA) )
		# get offset
		offset = 0
		if 1 in TA :
			offset = TA[1]
		if pos_tasks and neg_tasks :
			left = pos_tasks[0]
			right = neg_tasks[0]
			if comp_operator == '<' :
				return PrecedenceLax(task_left=left,task_right=right,offset=offset)
			elif comp_operator == '<=' :
				return PrecedenceTight(task_left=left,task_right=right,offset=offset)
			elif comp_operator == '<<' :
				return PrecedenceCond(task_left=left,task_right=right,offset=offset)
		elif pos_tasks and not neg_tasks:
			left = pos_tasks[0]
			right = -offset
			if comp_operator == '<':
				return BoundUp(task=left,bound=right)
			elif comp_operator == '<=':
				return BoundUpTight(task=left,bound=right)
			elif comp_operator == '>':
				return BoundLow(task=left,bound=right)
			elif comp_operator == '>=':
				return BoundLowTight(task=left,bound=right)
		elif neg_tasks and not pos_tasks:
			left = neg_tasks[0]
			right = offset
			if comp_operator == '<':
				return BoundLow(task=left,bound=right)
			elif comp_operator == '<=':
				return BoundLowTight(task=left,bound=right)
			elif comp_operator == '>':
				return BoundUp(task=left,bound=right)
			elif comp_operator == '>=':
				return BoundUpTight(task=left,bound=right)
		raise Exception('ERROR: sth is wrong')

	def __lt__(self,other) :
		if _isiterable(other) :
			return [ self < x for x in other ]
		if not isinstance(other,type(self)) :
			return self < _TaskAffine(other)
		return self._get_prec(self-other,'<')

	def __gt__(self,other) :
		if _isiterable(other) :
			return [ self > x for x in other ]
		if not isinstance(other,type(self)) :
			return self > _TaskAffine(other)
		return _TaskAffine(other) < self

	def __le__(self,other) :
		if _isiterable(other) :
			return [ self <= x for x in other ]
		if not isinstance(other,type(self)) :
			return self <= _TaskAffine(other)
		return self._get_prec(self-other,'<=')

	def __ge__(self,other) :
		if _isiterable(other) :
			return [ self >= x for x in other ]
		if not isinstance(other,type(self)) :
			return self >= _TaskAffine(other)
		return _TaskAffine(other) <= self

	def __lshift__(self,other) :
		if _isiterable(other) :
			return [ self << x for x in other ]
		if not isinstance(other,type(self)) :
			return self << _TaskAffine(other)
		return self._get_prec(self-other,'<<')

	def __rshift__(self,other) :
		if _isiterable(other) :
			return [ self >> x for x in other ]
		if not isinstance(other,type(self)) :
			return self >> _TaskAffine(other)
		return _TaskAffine(other) << self



class _Constraint(_SchedElement) :

	def __init__(self) :
		_SchedElement.__init__(self)

	def tasks(self):
		return list()

	def resources(self):
		return list()



class _Bound(_Constraint) :

	def __init__(self,task,bound) :
		_Constraint.__init__(self)
		self.task = task
		self.bound = bound
		self.comp_operator = '<'

	def tasks(self) :
		return [self.task]

	def __repr__(self) :
		return str(self.task) + ' ' + str(self.comp_operator) + ' ' + str(self.bound)

	def __str__(self) :
		return self.__repr__()

	def __hash__(self) :
		return self.__repr__().__hash__()



class BoundLow(_Bound) :
	"""
	A bound of the form T > 3
	"""
	def __init__(self,task,bound) :
		_Bound.__init__(self,task,bound)
		self.comp_operator = '>'



class BoundUp(_Bound) :
	"""
	A bound of the form T < 3
	"""
	def __init__(self,task,bound) :
		_Bound.__init__(self,task,bound)
		self.comp_operator = '<'



class BoundLowTight(_Bound) :
	"""
	A bound of the form T >= 3
	"""
	def __init__(self,task,bound) :
		_Bound.__init__(self,task,bound)
		self.comp_operator = '>='



class BoundUpTight(_Bound) :
	"""
	A bound of the form T <= 3
	"""
	def __init__(self,task,bound) :
		_Bound.__init__(self,task,bound)
		self.comp_operator = '<='



class _Precedence(_Constraint) :
	"""
	A precedence constraint of two tasks, left and right, and an offset.
	"""
	def __init__(self,task_left,task_right,offset=0) :
		_Constraint.__init__(self)
		self.task_left = task_left
		self.task_right = task_right
		self.offset = offset
		self.comp_operator = '<'

	def tasks(self) :
		return [self.task_left,self.task_right]

	def __repr__(self) :
		s = str(self.task_left) + ' '
		if self.offset > 0 :
			s += '+ ' + str(self.offset) + ' '
		s += str(self.comp_operator) + ' ' + str(self.task_right)
		if self.offset < 0 :
			s += ' + ' + str(-self.offset) + ' '
		return s

	def __str__(self) :
		return self.__repr__()

	def __hash__(self) :
		return self.__repr__().__hash__()



class PrecedenceLax(_Precedence) :
	"""
	A precedence of the form T1 + 3 < T2
	"""
	def __init__(self,task_left,task_right,offset=0) :
		_Precedence.__init__(self,task_left,task_right,offset)
		self.comp_operator = '<'



class PrecedenceTight(_Precedence) :
	"""
	A precedence of the form T1 + 3 <= T2
	"""
	def __init__(self,task_left,task_right,offset=0) :
		_Precedence.__init__(self,task_left,task_right,offset)
		self.comp_operator = '<='



class PrecedenceCond(_Precedence) :
	"""
	A precedence of the form T1 + 3 <= T2
	"""
	def __init__(self,task_left,task_right,offset=0) :
		_Precedence.__init__(self,task_left,task_right,offset)
		self.comp_operator = '<<'




class Resource(_SchedElement) :
	"""
	A resource which can processes tasks
	"""
	def __init__(self,name=None,size=1) :
		_SchedElement.__init__(self,name)
		self.size = size

	def __mul__(self,other) :
		return _ResourceAffine(self).__mul__(other)

	def __or__(self,other) :
		return _ResourceAffine(self) | other

	def __getitem__(self, key):
		C = _Capacity(resource=self)
		return C[key]

	def __le__(self,other):
		return _Capacity(resource=self) <= other

	def __ge__(self,other):
		return _Capacity(resource=self) >= other



class _ResourceAffine(_SchedElementAffine) :

	def __init__(self,unknown=None) :
		_SchedElementAffine.__init__(self,unknown=unknown,element_class=Resource,affine_operator='|')

	def __or__(self,other) :
		return super(_ResourceAffine,self).__add__(_ResourceAffine(other)) #add of superclass



class _Capacity(_Constraint):

	def __init__(self,resource):
		_Constraint.__init__(self)
		self.resource = resource
		self.weight = lambda T,t : 1
		self._param = 'length'
		self._start = None
		self._end = None
		self.bound = None
		self.comp_operator = None
		self.kind = 'sum'

	def __ge__(self, other):
		if not _isnumeric(other):
			raise Exception('ERROR: %s is not an integer, only integers are allowed'%str(other))
		self.__class__ = CapacityLow
		if 'diff' in self.kind:
			self.__class__ = CapacityDiffLow
		self.comp_operator = '>='
		self.bound = other
		return self

	def __le__(self, other):
		if not _isnumeric(other):
			raise Exception('ERROR: %s is not an integer, only integers are allowed'%str(other))
		self.__class__ = CapacityUp
		if 'diff' in self.kind:
			self.__class__ = CapacityDiffUp
		self.comp_operator = '<='
		self.bound = other
		return self

	def __getitem__(self,key):
		if isinstance(key,str):
			self._param = key
			weight_ = lambda T: T[self._param] if self._param in T else 0
		elif isinstance(key,int):
			weight_ = lambda t: 1 if t == key else 0
			self._start = key
			self._end = key+1
		elif isinstance(key,slice):
			self._start = key.start
			self._end = key.stop
			if self._start is not None and self._end is not None:
				weight_ = lambda t: 1 if t >= self._start and t < self._end else 0
			elif self._start is not None:
				weight_ = lambda t: 1 if t >= self._start else 0
			elif self._end is not None:
				weight_ = lambda t: 1 if t < self._end else 0
			else:
				weight_ = lambda t: 1
		elif hasattr(key, '__call__'):
			self._param = uuid.uuid4()
			weight_ = key
		else:
			raise Exception('ERROR: wrong slicing key for capacity')
		# concatenate slicing functions
		old_weight = copy.deepcopy(self.weight)
		weight_ = copy.deepcopy(weight_)
		if 'T' in weight_.__code__.co_varnames and 't' in weight_.__code__.co_varnames:
			new_weight = lambda T,t: old_weight(T,t)*weight_(T,t)
		elif 'T' in weight_.__code__.co_varnames:
			new_weight = lambda T,t=0: old_weight(T,t)*weight_(T)
		elif 't' in weight_.__code__.co_varnames:
			new_weight = lambda T,t: old_weight(T,t)*weight_(t)
		else:
			raise Exception('ERROR: wrong parameters in capacity selection function')
		self.weight = copy.deepcopy(new_weight)
		return self

	@property
	def diff(self):
		self.kind = 'diff'
		return self

	@property
	def dec(self):
		self.kind = 'diff_dec'
		return self

	@property
	def inc(self):
		self.kind = 'diff_inc'
		return self

	def __str__(self):
		param = self._param
		if self.name is not None and self.name is not '':
			param = self.name
		slice = ''
		if self._start is not None or self._end is not None:
			slice = '['
			if self._start is not None:
				slice += str(self._start)
			slice += ':'
			if self._end is not None: #large number
				slice += str(self._end)
			slice += ']'
		operator = ''
		if self.kind == 'diff':
			operator = '.diff()'
		elif self.kind == 'diff_inc':
			operator = '.inc()'
		elif self.kind == 'diff_dec':
			operator = '.dec()'
		s = '%s[\'%s\']%s%s %s %s' % (str(self.resource),str(param),slice,operator,
									  str(self.comp_operator),str(self.bound))
		return s

	def __repr__(self):
		return self.__str__()



class CapacityLow(_Capacity):
	"""
	A lower capacity bound on one resource in an interval
	"""
	def __init__(self,resource):
		_Capacity.__init__(self,resource)



class CapacityDiffLow(_Capacity):
	"""
	A lower bound on the number of on/off-switches for the capacity
	"""
	def __init__(self,resource):
		_Capacity.__init__(self,resource)



class CapacityUp(_Capacity):
	"""
	An upper capacity bound on one resource in an interval
	"""
	def __init__(self,resource):
		_Capacity.__init__(self,resource)



class CapacityDiffUp(_Capacity):
	"""
	An upper bound on the number of on/off-switches for the capacity
	"""
	def __init__(self,resource):
		_Capacity.__init__(self,resource)



import os
from time import clock
import re
import subprocess
import pulp
import pulp.solvers

class SCIP_CMD(pulp.solvers.LpSolver_CMD):
    def __init__(self, path = None, keepFiles = 0, mip = 1,
            msg = 0, options = [], time_limit = None, ratio_gap = None):
        pulp.solvers.LpSolver_CMD.__init__(self, path, keepFiles, mip, msg, options)
        self.time_limit = time_limit
        self.ratio_gap = ratio_gap

    """The SCIP LP solver"""
    def defaultPath(self):
        return self.executableExtension("scip")

    def available(self):
        """True if the solver is available"""
        return self.executable(self.path)

    def actualSolve(self, lp):
        """Solve a well formulated lp problem"""
        if not self.executable(self.path):
            raise pulp.PulpSolverError("PuLP: cannot execute "+self.path)
        if not self.keepFiles:
            pid = os.getpid()
            tmpLp = os.path.join(self.tmpDir, "%d-pulp.lp" % pid)
            tmpSol = os.path.join(self.tmpDir, "%d-pulp.sol" % pid)
        else:
            tmpLp = lp.name+"-pulp.lp"
            tmpSol = lp.name+"-pulp.sol"
        lp.writeLP(tmpLp, writeSOS = 0)
        #proc = ["scip", "-c", "read \"%s\"" % tmpLp, "-c", "set limits time 180", "-c", "optimize", "-c", "write solution \"%s\"" % tmpSol, "-c", "quit"]
        proc = ["scip", "-c", "read \"%s\"" % tmpLp]
        if self.time_limit is not None:
            proc += ["-c", "set limits time %f"%self.time_limit]
        if self.ratio_gap is not None:
            proc += ["-c", "set limits gap %f"%self.ratio_gap]
        proc += ["-c", "optimize", "-c", "write solution \"%s\"" % tmpSol, "-c", "quit"]
        proc.extend(self.options)

        self.solution_time = clock()
        if not self.msg:
            proc[0] = self.path
            pipe = open(os.devnull, 'w')
            rc = subprocess.call(proc, stdout = pipe,
                             stderr = pipe)
            if rc:
                raise pulp.PulpSolverError("PuLP: Error while trying to execute "+self.path)
        else:
            if os.name != 'nt':
                rc = os.spawnvp(os.P_WAIT, self.path, proc)
            else:
                rc = os.spawnv(os.P_WAIT, self.executable(self.path), proc)
            if rc == 127:
                raise pulp.PulpSolverError("PuLP: Error while trying to execute "+self.path)
        self.solution_time += clock()

        if not os.path.exists(tmpSol):
            raise pulp.PulpSolverError("PuLP: Error while executing "+self.path)
        lp.status, values = self.readsol(tmpSol)
        lp.assignVarsVals(values)
        if not self.keepFiles:
            try: os.remove(tmpLp)
            except: pass
            try: os.remove(tmpSol)
            except: pass
        return lp.status

    def readsol(self, filename):
        """Read a SCIP solution file"""
        with open(filename) as f:
            m = re.match(r"^solution status: (.*)", f.readline().strip())
            if not m:
                raise pulp.PulpSolverError("Unknown status returned by SCIP")
            statusString = m.group(1)
            scipStatus = {
                "unknown": pulp.LpStatusNotSolved,
                "user interrupt": pulp.LpStatusNotSolved,
                "node limit reached": pulp.LpStatusNotSolved,
                "total node limit reached": pulp.LpStatusNotSolved,
                "stall node limit reached": pulp.LpStatusNotSolved,
                "time limit reached": pulp.LpStatusOptimal,
                "memory limit reached": pulp.LpStatusNotSolved,
                "gap limit reached": pulp.LpStatusOptimal,
                "solution limit reached": pulp.LpStatusNotSolved,
                "solution improvement limit reached": pulp.LpStatusNotSolved,
                "optimal solution found": pulp.LpStatusOptimal,
                "infeasible": pulp.LpStatusInfeasible,
                "unbounded": pulp.LpStatusUnbounded,
                "infeasible or unbounded": pulp.LpStatusNotSolved,
                }
            if statusString not in scipStatus:
                raise pulp.PulpSolverError("Unknown status returned by SCIP")
            status = scipStatus[statusString]
            f.readline() # objective value:
            values = {}
            for line in f:
                name, val, _ = line.split()
                values[name] = float(val)
        return status, values

SCIP = SCIP_CMD



#! /usr/bin/env python
from __future__ import absolute_import
'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

import time, os, uuid


def _get_tmp_dir() :
	"""
	returns a default temporary directory
	"""
	if os.name != 'nt':
		# On unix use /tmp by default
		tmp_dir = os.environ.get("TMPDIR", "/tmp")
		tmp_dir = os.environ.get("TMP", tmp_dir)
	else:
		# On Windows use the current directory
		tmp_dir = os.environ.get("TMPDIR", "")
		tmp_dir = os.environ.get("TMP", tmp_dir)
		tmp_dir = os.environ.get("TEMP", tmp_dir)
	if not os.path.isdir(tmp_dir):
		tmp_dir = ""
	elif not os.access(tmp_dir, os.F_OK + os.W_OK):
		tmp_dif = ""
	return tmp_dir



def _get_dat_filename(scenario,msg=0) :

	S = scenario
	tmp_dir = _get_tmp_dir()
	unique_suffix = uuid.uuid4()
	dat_filename = tmp_dir+'/pyschedule_'+str(unique_suffix)+'.dat'

	resources = S.resources()
	tasks = S.tasks()

	id_to_resource = dict(zip(range(len(resources)),resources))
	resource_to_id = dict(zip(resources,range(len(resources))))

	id_to_task = dict(zip(range(len(tasks)),tasks))
	task_to_id = dict(zip(tasks,range(len(tasks))))

	LARGE_NUMBER = 1000

	# TODO: only upper capacity constraints
	Resources = [ (resource_to_id[R],0,LARGE_NUMBER) for R in resources if R.size == 1 ]
	Resources = sorted(Resources)

	CumulResources = [ (resource_to_id[R],R.size) for R in resources \
                           if  R.size > 1 ]

	# Tasks for .dat-file
	Tasks = list()
	TaskResources = list()
	TaskResourceGroups = list()
	TaskCumulResources = list()
	TaskTaskResources = list()
	for T in tasks :				
		Tasks.append( (task_to_id[T],int(T.length)) )
		single_resources = { R for RA in T.resources_req if len(RA) < 2 for R in RA }
		for RA in T.resources_req:
			# do not consider tasks which appear in a single resorce
			if len(RA) > 1 and set(RA) & set(single_resources):
				continue
			task_resource_group_id = T.resources_req.index(RA)
			for R in RA :
				task_id = task_to_id[T]
				resource_id = resource_to_id[R]
				if R.size == 1 :
					TaskResources.append( (task_id,resource_id,
					                      T.length,task_resource_group_id) )
					if (task_id,task_resource_group_id) not in TaskResourceGroups :
						TaskResourceGroups.append((task_id,task_resource_group_id))
				else :
					TaskCumulResources.append( (task_id,resource_id,RA[R]) )
	ra_to_tasks = S.resources_req_tasks()
	for RA in ra_to_tasks:
		tasks = list(ra_to_tasks[RA])
		T = tasks[0]
		for T_ in tasks[1:] :
			task_id_1 = task_to_id[T]
			task_id_2 = task_to_id[T_]
			for R in RA :
				resource_id = resource_to_id[R]
				TaskTaskResources.append((task_id_1,task_id_2,resource_id))

	Tasks = sorted(Tasks)
	TaskResources = sorted(TaskResources)
	TaskResourceGroups = sorted(TaskResourceGroups)
	TaskTaskResources = sorted(TaskTaskResources)

	# Precedences for .dat-file
	Precedences = list()
	for P in S.precs_lax() :
		if P.offset >= 0:
			Precedences.append(( task_to_id[P.task_left],task_to_id[P.task_right],P.offset))
		elif P.offset < 0:
			Precedences.append(( task_to_id[P.task_left],task_to_id[P.task_right],P.offset-P.task_left.length-P.task_right.length))

	TightPrecedences = list()
	for P in S.precs_tight() :
		if P.offset >= 0:
			TightPrecedences.append(( task_to_id[P.task_left],task_to_id[P.task_right],P.offset))
		elif P.offset < 0:
			TightPrecedences.append(( task_to_id[P.task_left],task_to_id[P.task_right],P.offset-P.task_left.length-P.task_right.length))

	CondPrecedences = list()
	for P in S.precs_cond() :
		CondPrecedences.append(( task_to_id[P.task_left],task_to_id[P.task_right],P.offset))

	LowerBounds = list()
	for P in S.bounds_low() :
		LowerBounds.append((task_to_id[P.task],P.bound))

	UpperBounds = list()
	for P in S.bounds_up() :
		UpperBounds.append((task_to_id[P.task],P.bound))
	if S.horizon is not None:
		for T in S.tasks():
			UpperBounds.append((task_to_id[T],S.horizon))

	FixBounds = list()
	for P in S.bounds_low_tight() :
		FixBounds.append((task_to_id[P.task],P.bound))
	for P in S.bounds_up_tight() :
		FixBounds.append((task_to_id[P.task],P.bound-P.task.length))
			
	# upper capacity bounds
	CapacityUps = list()
	CapacityUpTasks = list()
	CapacitySliceUps = list()
	CapacitySliceUpTasks = list()
	count = 0
	for C in S.capacity_up() :
		if C._start is None and C._end is None:
			CapacityUps.append((count,resource_to_id[C.resource],C.bound,-1,-1))
			for T in S.tasks() :
				if C.weight(T):
					CapacityUpTasks.append((count,task_to_id[T],C.weight(T)))
			'''
			else:
			start = C._start
			if start is None:
				start = 0
			end = C._end
			if end is None:
				end = S.horizon
			CapacitySliceUps.append((count,resource_to_id[C.resource],C.bound,start,end))
			'''
		count += 1

	# objective
	Objectives = [ (task_to_id[T],T.completion_time_cost) for T in S.tasks()
	              if 'completion_time_cost' in T ]
	Objectives = sorted(Objectives)

	# function to convert table into opl type typle lists
	def to_str(l) :
		return '\n'.join([ '<'+' '.join([ str(x) for x in row ])+'>'  \
                                  for row in l ])

	# write .dat-file
	with open(dat_filename,'w') as f :
		f.write('Objectives={\n'+to_str(Objectives)+'\n};\n\n')
		f.write('Tasks={\n'+to_str(Tasks)+'\n};\n\n')
		f.write('Resources={\n'+to_str(Resources)+'\n};\n\n')
		f.write('CumulResources={\n'+to_str(CumulResources)+'\n};\n\n')
		f.write('TaskResources={\n'+to_str(TaskResources)+'\n};\n\n')
		f.write('TaskResourceGroups={\n'+to_str(TaskResourceGroups)+'\n};\n\n')
		f.write('TaskCumulResources={\n'+to_str(TaskCumulResources)+'\n};\n\n')
		f.write('TaskTaskResources={\n'+to_str(TaskTaskResources)+'\n};\n\n')
		f.write('Precedences={\n'+to_str(Precedences)+'\n};\n\n')
		f.write('TightPrecedences={\n'+to_str(TightPrecedences)+'\n};\n\n')
		f.write('CondPrecedences={\n'+to_str(CondPrecedences)+'\n};\n\n')
		f.write('UpperBounds={\n'+to_str(UpperBounds)+'\n};\n\n')
		f.write('LowerBounds={\n'+to_str(LowerBounds)+'\n};\n\n')
		f.write('FixBounds={\n'+to_str(FixBounds)+'\n};\n\n')
		f.write('CapacityUps={\n'+to_str(CapacityUps)+'\n};\n\n')
		f.write('CapacitySliceUps={\n'+to_str(CapacitySliceUps)+'\n};\n\n')
		f.write('CapacityUpTasks={\n'+to_str(CapacityUpTasks)+'\n};\n\n')
		f.close()

	return dat_filename, task_to_id, id_to_resource



def _read_solution(scenario,log,task_to_id,id_to_resource,msg=0) :
	S = scenario

	# parse output 
	from pyparsing import Keyword,Literal,Word,alphas,nums,printables,OneOrMore,ZeroOrMore,dblQuotedString,Group
	INT = Word( nums )
	int_row = Group( INT + Literal(",").suppress() + \
                         INT + Literal(",").suppress() + \
			 INT + Literal(";").suppress() )
	plan = Group( Group( ZeroOrMore(int_row) ) )

	try:
		start_str, end_str = '##START_SOLUTION##', '##END_SOLUTION##'
		start_i, end_i = log.index(start_str)+len(start_str), log.index(end_str)
	except:
		print(log)
		if msg:
			print('ERROR: no solution found')
		return 0

	opl_plan = plan.parseString(log[start_i:end_i])
	int_plan = opl_plan[0][0]

	# get starts and resource assignments
	starts = dict()
	assign = dict()
	for row in int_plan :
		task_id = int(row[0])
		starts[task_id] = int(row[2])
		if task_id not in assign :
			assign[task_id] = list()
		assign[task_id].append(int(row[1]))

	# add to scenario
	for T in S.tasks() :
		T.start_value = starts[task_to_id[T]]
		if T.resources is None :
			T.resources = list()
		T.resources = [ id_to_resource[j] for j in assign[task_to_id[T]] ]

	return 1



def _get_mod_filename(mod_filename=None) :
	if mod_filename is None :
		solvers_path = os.path.dirname(os.path.realpath(__file__))
		return solvers_path+'/cpoptimizer.mod'
	return mod_filename



def solve(scenario,mod_filename=None,msg=0) :
	""" solve using cpoptimzer, make sure that the executable oplrun is in your PATH """
	S = scenario
	mod_filename = _get_mod_filename(mod_filename)
	dat_filename, task_to_id, id_to_resource = _get_dat_filename(scenario,msg=msg)

	# run cp-optimizer
	start_time = time.time()
	log = os.popen('oplrun %s %s' % (mod_filename, dat_filename) ).read()
	if msg :
		print('INFO: execution time (sec) = '+str(time.time()-start_time))

	# read solution
	return _read_solution(S,log,task_to_id,id_to_resource,msg=msg)



def solve_docloud(scenario,api_key,
                  base_url='https://api-oaas.docloud.ibmcloud.com/job_manager/rest/v1/',
                  mod_filename=None,msg=0) :
	""" solve using DOCloud, api_key is required """
	S = scenario
	mod_filename = _get_mod_filename(mod_filename)
	dat_filename, task_to_id, id_to_resource = _get_dat_filename(scenario,msg=msg)

	# solve and read solution
	from . import docloud
	log = docloud.solve(base_url=base_url,api_key=api_key,filenames=[mod_filename,dat_filename],msg=msg)
	if msg :
		print(log)
	return _read_solution(S,log,task_to_id,id_to_resource,msg=msg)








#! /usr/bin/env python
from __future__ import print_function

'''
Copyright 2015 Tim Nonner
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

import copy

def sort_with_precs(scenario) :
	"""
	returns the tasks of the given scenario sorted according to the
	lax precedence constraints and next according to the task length,
	where large tasks have priority
	"""
	S = scenario

	'''
	# get direct children for topological sort
	children = { T : set() for T in S.tasks() }
	for P in S.precs_lax() :
		children[P.left].add(P.right)
	# extend children to other descendents
	for T in S.tasks() :
		task_stack = {T}
		while task_stack :
			T_ = task_stack.pop()
			children[T] = children[T] | children[T_]
			task_stack = task_stack | children[T_]
	'''

	import networkx as nx
	G = nx.DiGraph()
	G.add_nodes_from(S.tasks())
	G.add_edges_from([ (P.task_left,P.task_right) for P in S.precs_lax() ])
	task_list = nx.algorithms.topological_sort(G)
	return task_list


#TODO: list as parameter of solving procedure
def solve(scenario,solve_method,task_list=None,batch_size=1,plot_method=None,msg=0) :
	"""
	Iteratively adds tasks and uses solve_method to integrate these
	tasks into the schedule.

	Arguments:
		scenario   : the scenario to solve
		task_list  : list of all tasks which defines the order in which all tasks are
                     added to the schedule
		batch_size : the number of tasks to integrate in the schedule at a time
	"""

	S = scenario

	if task_list is None :
		task_list = sort_with_precs(S)

	constraints = S._constraints # keep references and clear old reference list
	S._constraints = []

	#non_objective_tasks = [ T for T in task_list if not T.objective ]
	for T in task_list :
		S -= T #remove all tasks which are not part of objective

	def batches(tasks, batch_size):
		for i in xrange(0, len(tasks), batch_size):
			yield tasks[i:i+batch_size]

	for batch in batches(task_list,batch_size) :
		if msg :
			print('INFO: batch for list scheduling '+','.join([ str(T) for T in batch]))
		for T in batch :
			S += T
		S._constraints = [ C for C in constraints if set(C.tasks()).issubset(set(S.tasks())) ]

		solve_method(S)
		if plot_method is not None:
			plot_method(S)

		for T in S.tasks():
			S += T >= T.start_value


		
		
		


		
		

		

	






#! /usr/bin/python
from __future__ import absolute_import as _absolute_import
from __future__ import print_function

'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''


"""
a simple interface to pulp which is also an interface blueprint for potential other mip-interfaces
"""

import pulp as pl
import time

from pyschedule.solvers.pulp_scip import SCIP_CMD

class MIP(object):
	"""
	Interface to pulp mip solver
	"""

	def __init__(self,name,kind='Minimize'):
		kinds = {'Minimize':pl.LpMinimize, 'Maximize':pl.LpMaximize}
		# self.mip = pl.LpProblem(name, kinds[kind])
		self.mip = pl.LpProblem('Integer Program', kinds[kind])

	def var(self,name,low=0,up=0,cat='Binary'):
		return pl.LpVariable(name, low, up, cat=cat)

	def con(self,affine,sense=0,rhs=0):
		con = pl.LpConstraint(pl.LpAffineExpression(affine),sense=sense,rhs=rhs)
		self.mip += con
		return con

	def obj(self,affine):
		self.mip += pl.LpAffineExpression(affine)

	def solve(self,msg=0,**kwarg):
		# kind = 'CBC'
		if 'kind' in kwarg:
			kind = kwarg['kind']
		time_limit = None
		if 'time_limit' in kwarg:
			time_limit = float(kwarg['time_limit'])
		random_seed = None
		if 'random_seed' in kwarg:
			random_seed = kwarg['random_seed']
		ratio_gap = None
		if 'ratio_gap' in kwarg:
			ratio_gap = float(kwarg['ratio_gap'])
		start_time = time.time()
		# select solver for pl
		if kind == 'CPLEX':
			if time_limit is not None:
				# pulp does currently not support a timelimit in 1.5.9
				self.mip.solve(pl.CPLEX_CMD(msg=msg, timelimit=time_limit))
			else:
				self.mip.solve(pl.CPLEX_CMD(msg=msg))
		elif kind == 'GLPK':
			self.mip.solve(pl.GLPK_CMD(msg=msg))
		elif kind == 'SCIP':
			self.mip.solve(SCIP_CMD(msg=msg,time_limit=time_limit,ratio_gap=ratio_gap))
		elif kind == 'CBC' or kind == 'COIN':
			options = []
			if time_limit is not None:
				options.extend(['sec', str(time_limit)])
			if random_seed is not None:
				options.extend(['randomSeed', str(random_seed)])
				options.extend(['randomCbcSeed', str(random_seed)])
			if ratio_gap is not None:
				options.extend(['ratio', str(ratio_gap)])
			if kind == 'CBC':
				self.mip.solve(pl.PULP_CBC_CMD(msg=msg, options=options))
			elif kind == 'COIN':
				self.mip.solve(pl.COIN(msg=msg, options=options))
		else:
			raise Exception('ERROR: solver ' + kind + ' not known')

		if msg:
			print('INFO: execution time for solving mip (sec) = ' + str(time.time() - start_time))
		if self.mip.status == 1 and msg:
			print('INFO: objective = ' + str(pl.value(self.mip.objective)))

	def status(self):
		return self.mip.status

	def value(self,var):
		return var.varValue



#! /usr/bin/env python
from __future__ import absolute_import as _absolute_import

'''
Copyright 2015 Tim Nonner
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

import time, os, copy, collections



def solve(scenario,time_limit=None,copy_scenario=False,msg=0) :
	"""
	Integration of the ortools scheduling solver
	"""

	try :
		from ortools.constraint_solver import pywrapcp
	except :
		raise Exception('ERROR: ortools is not installed')

	S = scenario
	if copy_scenario :
		S = copy.deepcopy(scenario)

	ort_solver = pywrapcp.Solver(S.name)

	# tasks
	task_to_interval = collections.OrderedDict()
	resource_to_intervals = resource_to_intervals = { R : list() for R in S.resources() }
	resource_task_to_interval = collections.OrderedDict()
	for T in S.tasks() :
		I = ort_solver.FixedDurationIntervalVar(0,S.horizon-T.length,T.length,False,T.name)
		task_to_interval[T] = I




	# resourcee requirements
	for T in S.tasks():
		I = task_to_interval[T]
		for RA in T.resources_req:
			RA_tasks = list()
			for R in RA :
				I_ = ort_solver.FixedDurationIntervalVar(0,S.horizon-T.length,T.length,True,T.name+'_'+R.name)
				resource_to_intervals[R].append(I_)
				RA_tasks.append(I_)
				resource_task_to_interval[(R,T)] = I_
				ort_solver.Add( I.StaysInSync(I_) )
				# if resources are fixed
				if T.resources is not None and R in T.resources :
					ort_solver.Add( I_.PerformedExpr() == 1 )
			# one resource needs to get selected
			ort_solver.Add(ort_solver.Sum([ I_.PerformedExpr() for I_ in RA_tasks ]) == 1)
	ra_to_tasks = S.resources_req_tasks()
	for RA in ra_to_tasks:
		tasks = list(ra_to_tasks[RA])
		T = tasks[0]
		for T_ in tasks[1:] :
			for R in RA :
				I  = resource_task_to_interval[(R,T)]
				I_ = resource_task_to_interval[(R,T_)]
				ort_solver.Add( I.PerformedExpr() == I_.PerformedExpr() )

	# resources
	sequences = collections.OrderedDict()
	for R in S.resources() :
		disj = ort_solver.DisjunctiveConstraint(resource_to_intervals[R],R.name)
		sequences[R] = disj.SequenceVar()
		ort_solver.Add(disj)

	# move objective
	# TODO: bug, variables that are not part of the objective might not be finally defined
	ort_objective_var = ort_solver.Sum([ task_to_interval[T].EndExpr()*T.completion_time_cost
										 for T in S.tasks() if T in task_to_interval
										 and 'completion_time_cost' in T ])
	ort_objective = ort_solver.Minimize(ort_objective_var, 1)

	# precedences lax
	for P in S.precs_lax() :
		ort_solver.Add( task_to_interval[P.task_right].StartsAfterEnd( task_to_interval[P.task_left] ) )
		# TODO: add offset, but this requires DependecyGraph which is not exposed via swig?

	# precedences tight
	for P in S.precs_tight() :
		ort_solver.Add( task_to_interval[P.task_right].StartsAtEnd( task_to_interval[P.task_left] ) )
		# TODO: add offset, but this requires DependecyGraph which is not exposed via swig?

	# bound low
	for P in S.bounds_low() :
		ort_solver.Add( task_to_interval[P.task].StartsAfter(P.bound) )

	# bound up
	for P in S.bounds_up() :
		ort_solver.Add( task_to_interval[P.task].StartsBefore(P.bound) )

	# tight bound low
	for P in S.bounds_low_tight() :
		ort_solver.Add( task_to_interval[P.task].StartsAt(P.bound) )

	# tight bound up
	for P in S.bounds_up_tight() :
		ort_solver.Add( task_to_interval[P.task].EndsAt(P.bound) )

	# capacity lower bounds
	for C in S.capacity_low():
		# ignore sliced capacity constraints
		if C._start != None or C._end != None:
			continue
		R = C.resource
		cap_tasks = [ (resource_task_to_interval[R,T],C.weight(T=T,t=0)) for T in S.tasks() ]
		ort_solver.Add( ort_solver.Sum([ I.PerformedExpr()*w for I,w in cap_tasks ]) >= C.bound )

	# capacity lower bounds
	for C in S.capacity_low():
		# ignore sliced capacity constraints
		if C._start != None or C._end != None:
			continue
		R = C.resource
		cap_tasks = [ (resource_task_to_interval[R,T],C.weight(T=T,t=0)) for T in S.tasks() ]
		ort_solver.Add( ort_solver.Sum([ I.PerformedExpr()*w for I,w in cap_tasks ]) <= C.bound )

	# creates search phases.
	vars_phase = ort_solver.Phase([ort_objective_var],
		            ort_solver.CHOOSE_FIRST_UNBOUND,
		            ort_solver.ASSIGN_MIN_VALUE)
	sequence_phase = ort_solver.Phase(sequences.values(),
		                ort_solver.SEQUENCE_DEFAULT)
	main_phase = ort_solver.Compose([sequence_phase, vars_phase])

	# creates the search log.
	search_log = ort_solver.SearchLog(100, ort_objective_var)

	# collect solution
	solution = ort_solver.Assignment()
	for T in S.tasks() :
		solution.Add(task_to_interval[T])
	for R in S.resources() :
		for I in resource_to_intervals[R] :
			solution.Add(I)
	collector = ort_solver.LastSolutionCollector(solution)

	# set limits (time limit im ms)
	if time_limit :
		ort_time_limit = int(time_limit*1000)
	else :
		ort_time_limit = 100000000
	branch_limit = 100000000	
	failures_limit = 100000000
	solutions_limit = 10000000
	limits = (ort_solver.Limit(ort_time_limit, branch_limit, failures_limit, solutions_limit, True))

	# add log if mst is requested
	search_params = [limits,collector,ort_objective]
	if msg :
		search_params.append(search_log)

	# solves the problem.
	ort_solver.Solve(main_phase,search_params)

	# check for a solution
	if not collector.SolutionCount():
		if msg:
			print('ERROR: no solution found')
		return 0
	solution = collector.Solution(0)

	# read last solution
	for T in S.tasks() :
		T.start_value = int(solution.StartMin(task_to_interval[T])) #collector.StartValue(0, task_to_interval[T])
		T.resources = [ R \
	                    for RA in T.resources_req for R in RA \
	                    if collector.PerformedValue(0,resource_task_to_interval[(R,T)]) == 1 ]
	return 1
	





































#! /usr/bin/env python
'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

"""
This module contains solvers that accept the pyschedule scenario format
as input. Each solver should correspond to a class which offers the
method "solve" that takes a scenario as minimal input. If a solution
exists, then a 0 should be returned, otherwise a 1. The optimal solution
should be directly written to the passed scenario
"""

from . import mip
from . import cpoptimizer
from . import ortools
from . import listsched



from __future__ import absolute_import as _absolute_import
import operator

#! /usr/bin/env python
'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''


def plot(scenario,img_filename=None,resource_height=1.0,show_task_labels=True,
         color_prec_groups=False,hide_tasks=[],hide_resources=[],task_colors=dict(),fig_size=(15,5)) :
	"""
	Plot the given solved scenario using matplotlib

	Args:
		scenario:    scenario to plot
		msg:         0 means no feedback (default) during computation, 1 means feedback
	"""
	try :
		import matplotlib.patches as patches, matplotlib.pyplot as plt
	except :
		raise Exception('ERROR: matplotlib is not installed')
	import random

	S = scenario
	# trivial connected components implementation to avoid
	# having to import other packages just for that
	def get_connected_components(edges) :
		comps = dict()
		for v,u in edges :
			if v not in comps and u not in comps :
				comps[v] = v
				comps[u] = v
			elif v in comps and u not in comps :
				comps[u] = comps[v]
			elif v not in comps and u in comps :
				comps[v] = comps[u]
			elif v in comps and u in comps and comps[v] != comps[u] :
				old_comp = comps[u]
				for w in comps :
					if comps[w] == old_comp :
						comps[w] = comps[v]
		# replace component identifiers by integers startting with 0
		values = list(comps.values())
		comps = { T : values.index(comps[T]) for T in comps }
		return comps

	tasks = [ T for T in S.tasks() if T not in hide_tasks ]
	
	# get connected components dict for coloring
	# each task is mapping to an integer number which corresponds
	# to its connected component
	edges = [ (T,T) for T in tasks ]
	if color_prec_groups :
		edges += [ (T,T_) for P in set(S.precs_lax()) | set(S.precs_tight()) \
	                   for T in P.tasks() for T_ in P.tasks() \
                           if T in tasks and T_ in tasks ]
	comps = get_connected_components(edges)

	# color map
	colors = ['#7EA7D8','#A1D372','#EB4845','#7BCDC8','#FFF79A'] #pastel colors
	#colors = ['red','green','blue','yellow','orange','black','purple'] #basic colors
	colors += [ [ random.random() for i in range(3) ] for x in range(len(S.tasks())) ] #random colors
	color_map = { T : colors[comps[T]] for T in comps }
	# replace colors with fixed task colors

	for T in task_colors :
		color_map[T] = task_colors[T]

	solution = S.solution()
	hide_tasks_str = [ T for T in hide_tasks ]
	solution = [ (T,R,x,y) for (T,R,x,y) in solution if T not in hide_tasks_str ] #tasks of zero length are not plotted

	fig, ax = plt.subplots(1, 1, figsize=fig_size)
	resource_sizes_count = 0
	visible_resources = [ R for R in S.resources() if R not in hide_resources ]
	if not visible_resources:
		raise Exception('ERROR: no resources to plot')
	total_resource_sizes = sum([ R.size for R in visible_resources ])
	R_ticks = list()

	for R in visible_resources :
		if R.size is not None :
			resource_size = R.size
		else :
			resource_size = 1.0
		R_ticks += [str(R.name)]*int(resource_size)
		# compute the levels of the tasks on one resource
		task_levels = dict()
		# get solution on resource sorted according to start time
		R_solution = [ (T,R_,x,y) for (T,R_,x,y) in solution if R_ == R ]
		R_solution = sorted(R_solution, key=lambda x : x[2])
		# iteratively fill all levels on resource, start with empty fill
		level_fill = { i : 0 for i in range(int(resource_size)) }
		for T,R_,x,y in R_solution :
			sorted_levels =  sorted(level_fill.items(), key = operator.itemgetter(1, 0))
			# get the maximum resource requirement
			coeff = max([ RA[R] for RA in T.resources_req if R_ in RA ])
			min_levels = [ level for level,fill in sorted_levels[:coeff] ]
			task_levels[T] = min_levels
			for level in min_levels :
				level_fill[level] += T.length
		# plot solution
		for (T,R,x,x_) in R_solution :
			for level in task_levels[T] :
				y = (total_resource_sizes-1-(resource_sizes_count+level))*resource_height
				ax.add_patch(
				    patches.Rectangle(
					(x, y),       # (x,y)
					max(x_-x,0.5),   # width
					resource_height,   # height
					color = color_map[T],
					alpha=0.6
				    )
				)
				if show_task_labels :
					plt.text(x,y+0.1*resource_height,str(T.name),fontsize=14,color='black')
		resource_sizes_count += resource_size

	# format graph
	plt.title(str(S.name))
	plt.yticks([ resource_height*x + resource_height/2.0 for x in range(len(R_ticks)) ],R_ticks[::-1])
	plt.ylim(0,resource_sizes_count*resource_height)#resource_height*len(resources))
	plt.xlim(0,max([ x_ for (I,R,x,x_) in solution if R in visible_resources ]))
	if img_filename :
		fig.figsize=(1,1)
		plt.savefig(img_filename,dpi=200,bbox_inches='tight')
	else :
		plt.show()





from setuptools import setup, find_packages

setup(
    name='pyschedule',
    version='0.2.17',
    description='A python package to formulate and solve resource-constrained scheduling problems',
    url='https://github.com/timnon/pyschedule',
    author='Tim Nonner',
    author_email='tim@nonner.de',
    license='Apache 2.0',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['pulp']
)



#!/usr/bin/python
# coding: utf-8
from __future__ import print_function

'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

import json, requests, os, time, sys

class _DOcloud:
	"""
	Client for DOcloud
	Args:
		base_url:	    base url of DOcloud, get at https://developer.ibm.com/docloud/
		api_key:	     api key, get at https://developer.ibm.com/docloud/
		msg:		 0 means no feedback (default) during computation, 1 means feedback
	"""
    
	def __init__(self, base_url, api_key, msg=0):
		self.url = '%sjobs' % base_url
		self.auth = {'X-IBM-Client-ID' : api_key}
		self.headers_post = {'X-IBM-Client-ID' : api_key, 'Content-Type' : 'application/json'}
		self.headers_put = {'X-IBM-Client-ID' : api_key, 'Content-Type' : 'application/octet-stream'}
		self.msg = msg
		requests.delete(self.url, headers=self.auth)
	
	def execute(self, filenames) :
		# get files and add to attachments
		mydata = {}
		for filename in filenames:
			filename_short = os.path.basename(filename)
			with open(filename) as f:
				mydata[filename_short] = f.read()    

		# create job with file list
		attachments = {'attachments' : list(mydata.keys())}
		r = requests.post(self.url, headers=self.headers_post, data=json.dumps(attachments))
		if r.status_code != 201 :
			print('ERROR: unsuccessful job creation %s' % r.status_code)
		job = r.headers['location']

		# send files
		for filename in mydata :
			r = requests.put('%s/attachments/%s/blob' % (job,filename), 
					 headers=self.headers_put, data=mydata[filename])
			if r.status_code != 204 :
				print('ERROR: %s upload failed %d' % (filename, r.status_code))
			elif self.msg :
					print('INFO: %s upload succeeded' % filename)

		# execute job
		r = requests.post(job+'/execute', headers=self.auth)
		if r.status_code != 204 :
			print('ERROR: execute failed to start %d' % r.status_code)
		elif self.msg :
			print('INFO: cloud execution successfull')
		return job
    
	def get_info(self, job) :
		r = requests.get(job, headers=self.headers_post)
		if r.status_code != 200:
			print('ERROR: cannot get info %s' % r.status_code)
		elif self.msg :
			print('INFO: info downloaded succesfully')
		return json.loads(r.content.decode('utf-8'))
    
	def get_log(self, job) :
		r = requests.get(job+'/log/blob', headers=self.headers_post)
		if r.status_code != 200:
			print('ERROR: cannot get info %s' % r.status_code)
		elif self.msg :
			print('INFO: log downloaded successfully')
		return r.content.decode('utf-8')

	def get_result(self, job) :
		status = 'RUNNING'
		job_info = {}
		# wait until job is finished
		while status != 'PROCESSED' and status != 'FAILED':
			time.sleep(1)
			job_info = self.get_info(job)
			status = job_info['executionStatus']
			if self.msg:
				print('INFO: job status : %s' % status)
		if status == 'FAILED' :
			print('ERROR: job failed : %s' % job_info['failure'])
			return {}
		else:
			r = requests.get('%s/attachments/solution.json/blob' % job, headers=self.headers_post)
			if r.status_code != 200:
				print('ERROR: result download failed %d' % r.status_code)
			elif self.msg :
				print('INFO: result downloaded successfully')
			return json.loads(r.text)
    
	def clean(self, job) :
		requests.delete(job, headers=self.auth)
	
	def solve(self, filenames) :
		""" send files with names in filenames to DOcloud and solve """

		job = self.execute(filenames)
		result = self.get_result(job)
		log = self.get_log(job)
		self.clean(job)
		return log



def solve(base_url,api_key,filenames,msg=0) :
	"""
	solve the model in the given files using DOcloud. Returns the log-file, so write the solution there
	Args:
		base_url:	 base url of DOcloud, get at https://developer.ibm.com/docloud/
		api_key:	 api key, get at https://developer.ibm.com/docloud/
		msg:		 0 means no feedback (default) during computation, 1 means feedback		
	"""
	try :
		import OpenSSL
	except :
		raise Exception('ERROR: pyOpenSSL needs to be installed to use docloud (or python 2.7.9 above). \
                                        Also install ndg-httpsclient and pyasn1')
	doc = _DOcloud(base_url, api_key, msg=msg)
	log = doc.solve(filenames=filenames)
	return log
	









#! /usr/bin/python
from __future__ import absolute_import as _absolute_import
from __future__ import print_function

'''
Copyright 2015 Tim Nonner

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''

from .mip_pulp import MIP
import collections



def _isnumeric(var):
	return isinstance(var, (int))  # only integers are accepted

def _get_task_groups(scenario):
	"""
	computes the task groups according to the _task_group parameter
	returns a mapping a task group representative to the task group
	"""
	_task_groups = collections.OrderedDict()
	for T in scenario.tasks():
		if 'group' in T:
			task_group_name = T.group
			if task_group_name in _task_groups:
				_task_groups[task_group_name].append(T)
			else:
				_task_groups[task_group_name] = [T]
	task_groups = collections.OrderedDict([ (_task_groups[task_group_name][0],
											 _task_groups[task_group_name])
										  for task_group_name in _task_groups])
	tasks_in_task_groups = [ T_ for T in task_groups for T_ in task_groups[T] ]
	task_groups.update([ (T,[T]) for T in scenario.tasks()
								 if T not in tasks_in_task_groups ])
	return task_groups

def solve(scenario, kind='CBC', time_limit=None, random_seed=None, ratio_gap=0.0, msg=0):
	"""
	Solves the given scenario using a discrete MIP

	Args:
		scenario:            scenario to solve
		kind:                MIP-solver to use: CPLEX, GLPK, CBC, SCIP
		time_limit:          a time limit, only for CPLEX, CBC and SCIP
		random_seed:         random seed
		ratio_gap:           MIP-gap
		msg:                 0 means no feedback (default) during computation, 1 means feedback

	Returns:
		1 if solving was successful
		0 if solving was not successful
	"""
	scenario.check()
	mip = MIP(str(scenario))
	return DiscreteMIP(mip).solve(scenario, kind=kind, time_limit=time_limit, random_seed=random_seed, ratio_gap=ratio_gap, msg=msg)

def solve_bigm(scenario, bigm=10000, kind='CBC', time_limit=None, random_seed=None, ratio_gap=0.0, msg=0):
	"""
	Solves the given scenario using a bigm-type MIP

	Args:
		scenario:    scenario to solve
		kind:        MIP-solver to use: CPLEX, GLPK, CBC or SCIP
		bigm :       a large number to allow a big-m type model
		time_limit:  a time limit, only for CPLEX, CBC and SCIP
		random_seed: random_seed
		ratio_gap:   MIP-gap
		msg:         0 means no feedback (default) during computation, 1 means feedback

	Returns:
		scenario is solving was successful
		None if solving was not successful
	"""
	scenario.check()
	mip = MIP(str(scenario))
	return ContinuousMIP(mip).solve(scenario, bigm=bigm, kind=kind, time_limit=time_limit, random_seed=random_seed,
									ratio_gap=ratio_gap, msg=msg)



class ContinuousMIP(object):
	"""
	An interface to the pulp MIP solver package, supported are CPLEX, GLPK, CBC
	"""

	def __init__(self,mip):
		self.mip = mip
		self.scenario = None
		self.horizon = None
		self.bigm = None
		self.x = None  # mip variables shortcut


	def build_mip_from_scenario(self, task_groups=None, msg=0):

		S = self.scenario
		BIGM = self.bigm

		mip = self.mip
		#mip = pl.LpProblem(str(S), pl.LpMinimize)
		cons = list() # log of constraints for debugging

		# task variables
		x = dict()

		for T in S.tasks():
			x[T] = mip.var(str(T),up=self.horizon,cat='Continuous')#pl.LpVariable(str(T), 0)
			# add task vs resource variabls
			for RA in T.resources_req:
				for R in RA:
					x[(T, R)] = mip.var(str((T, R))) #pl.LpVariable(str((T, R)), 0, 1, cat=pl.LpBinary)

		# resources req
		for T in S.tasks():
			for RA in T.resources_req:
				if len(RA) > 1:
					continue
				affine = [ (x[(T, R)], 1) for R in RA ]
				cons.append(mip.con(affine,sense=0,rhs=1))
				#mip += sum([x[(T, R)] for R in RA]) == 1
		ra_to_tasks = S.resources_req_tasks()
		for RA in ra_to_tasks:
			tasks = list(ra_to_tasks[RA])
			T = tasks[0]
			affine = [ (x[(T, R)], 1) for R in RA ]
			cons.append(mip.con(affine,sense=0,rhs=1))
			#mip += sum([x[(T, R)] for R in RA]) == 1
			for T_ in tasks[1:]:
				for R in RA:
					affine = [ (x[(T, R)],1), (x[(T_, R)],-1) ]
					cons.append(mip.con(affine,sense=0,rhs=0))
					#mip += x[(T, R)] == x[(T_, R)]

		# objective
		affine = [ (x[T], T.completion_time_cost) for T in S.tasks()
				   if 'completion_time_cost' in T and T in x ]
		mip.obj(affine)
		#mip += pl.LpAffineExpression([ (x[T], T.completion_time_cost) for T in S.tasks()
		#                                        if 'completion_time_cost' in T and T in x ])

		# same resource variable
		task_pairs = [(T, T_) for T in S.tasks() for T_ in S.tasks() if str(T) < str(T_)]
		for (T, T_) in task_pairs:
			if T.resources:
				resources = T.resources
			else:
				resources = S.resources(task=T)
			if T_.resources:
				resources_ = T_.resources
			else:
				resources_ = S.resources(task=T_)
			shared_resources = list(set(resources) & set(resources_))

			# TODO: restrict the number of variables
			if shared_resources:
				x[(T, T_, 'SameResource')] = \
					mip.var(str((T, T_, 'SameResource')),up=S.horizon,cat='Integer')
				#   pl.LpVariable(str((T, T_, 'SameResource')), lowBound=0)  # ,cat=pl.LpInteger)
				x[(T_, T, 'SameResource')] = \
					mip.var(str((T_, T, 'SameResource')),up=S.horizon,cat='Integer')
				#   pl.LpVariable(str((T_, T, 'SameResource')), lowBound=0)  # ,cat=pl.LpInteger)
				affine = [ (x[(T, T_, 'SameResource')],1), (x[(T_, T, 'SameResource')],-1) ]
				cons.append(mip.con(affine,sense=0,rhs=0))
				#mip += x[(T, T_, 'SameResource')] == x[(T_, T, 'SameResource')]
				for R in shared_resources:
					affine = [ (x[(T, R)],1), (x[(T_, R)],1), (x[(T, T_, 'SameResource')],-1)]
					cons.append(mip.con(affine,sense=-1,rhs=1))
					#mip += x[(T, R)] + x[(T_, R)] - 1 <= x[(T, T_, 'SameResource')]
				# ordering variables
				x[(T, T_)] = mip.var(str((T, T_)))#pl.LpVariable(str((T, T_)), 0, 1, cat=pl.LpBinary)
				x[(T_, T)] = mip.var(str((T_, T)))#pl.LpVariable(str((T_, T)), 0, 1, cat=pl.LpBinary)
				affine = [ (x[(T, T_)],1), (x[(T_, T)],1)]
				cons.append(mip.con(affine,sense=0,rhs=1))
				#mip += x[(T, T_)] + x[(T_, T)] == 1

				affine = [ (x[T],1), (x[T_],-1), (x[(T, T_)],BIGM), (x[(T, T_, 'SameResource')], BIGM) ]
				cons.append(mip.con(affine,sense=-1,rhs=2*BIGM-T.length))
				#mip += x[T] + T.length <= x[T_] + \
				#           (1 - x[(T, T_)]) * BIGM + (1 - x[(T, T_, 'SameResource')]) * BIGM
				affine = [ (x[T_],1), (x[T],-1), (x[(T, T_)],-BIGM), (x[(T, T_, 'SameResource')], BIGM) ]
				cons.append(mip.con(affine,sense=-1,rhs=BIGM-T_.length))
				#mip += x[T_] + T_.length <= x[T] + \
				#        x[(T, T_)] * BIGM + (1 - x[(T, T_, 'SameResource')]) * BIGM


		# precedence constraints
		for P in S.precs_lax():
			if P.offset >= 0:
				affine = [ (x[P.task_left],1), (x[P.task_right],-1) ]
				cons.append(mip.con(affine,sense=-1,rhs=-P.task_left.length-P.offset))
				#mip += x[P.task_left] + P.task_left.length + P.offset <= x[P.task_right]
			elif P.offset < 0:
				affine = [ (x[P.task_left],1), (x[P.task_right],-1) ]
				cons.append(mip.con(affine,sense=-1,rhs=P.task_right.length-P.offset))
				#mip += x[P.task_left] <= x[P.task_right] + P.task_right.length - P.offset

		# tight precedence constraints
		for P in S.precs_tight():
			if P.offset >= 0:
				affine = [ (x[P.task_left],1), (x[P.task_right],-1) ]
				cons.append(mip.con(affine,sense=0,rhs=-P.task_left.length-P.offset))
				#mip += x[P.task_left] + P.task_left.length + P.offset == x[P.task_right]
			elif P.offset < 0:
				affine = [ (x[P.task_left],1), (x[P.task_right],-1) ]
				cons.append(mip.con(affine,sense=0,rhs=P.task_right.length-P.offset))
				#mip += x[P.task_left] == x[P.task_right] + P.task_right.length - P.offset

		# conditional precedence constraints
		for P in S.precs_cond():
			affine = [ (x[P.task_left],1), (x[P.task_right],-1), (x[(P.task_left, P.task_right)],BIGM),
					   (x[(P.task_left, P.task_right, 'SameResource')],BIGM) ]
			cons.append(mip.con(affine,sense=-1,rhs=-P.task_left.length-P.offset+2*BIGM))
			#mip += x[P.task_left] + P.task_left.length + P.offset <= x[P.task_right] + \
			#                   (1 - x[(P.task_left, P.task_right)]) * BIGM + (1 - x[
			#   (P.task_left, P.task_right, 'SameResource')]) * BIGM

		# upper bounds
		for P in S.bounds_up():
			affine = [ (x[P.task],1) ]
			cons.append(mip.con(affine,sense=-1,rhs=P.bound-P.task.length))
			#mip += x[P.task] + P.task.length <= P.bound

		# lower bounds
		for P in S.bounds_low():
			affine = [ (x[P.task],1) ]
			cons.append(mip.con(affine,sense=1,rhs=P.bound))
			#mip += x[P.task] >= P.bound

		# tight upper bounds
		for P in S.bounds_up_tight():
			affine = [ (x[P.task],1) ]
			cons.append(mip.con(affine,sense=0,rhs=P.bound-P.task.length))
			#mip += x[P.task] + P.task.length == P.bound

		# tight lower bounds
		for P in S.bounds_low_tight():
			affine = [ (x[P.task],1) ]
			cons.append(mip.con(affine,sense=0,rhs=P.bound))
			#mip += x[P.task] == P.bound

		# capacity lower bounds
		for C in S.capacity_low():
			# ignore sliced capacity constraints
			if C._start != None or C._end != None:
				continue
			affine = [ (x[T, C.resource], C.weight(T=T,t=0)) for T in S.tasks()
					   if (T,C.resource) in x and C.weight(T=T,t=0) ]
			if not affine:
				continue
			cons.append(mip.con(affine, sense=1, rhs=C.bound))

		# capacity upper bounds
		for C in S.capacity_up():
			# ignore sliced capacity constraints
			if C._start != None or C._end != None:
				continue
			affine = [ (x[T, C.resource], C.weight(T=T,t=0)) for T in S.tasks()
					   if (T,C.resource) in x and C.weight(T=T,t=0) ]
			if not affine:
				continue
			cons.append(mip.con(affine, sense=-1, rhs=C.bound))

		'''
		for con in cons:
			mip.add_con(con)
		'''

		self.mip = mip
		self.x = x

	def read_solution_from_mip(self, msg=0):
		for T in self.scenario.tasks():
			if self.x[T].varValue is None:
				T.start_value = 0
			else:
				T.start_value = int(self.x[T].varValue)
			if T.resources:
				resources = T.resources
			else:
				resources = self.scenario.resources(task=T)

			task_resources = []
			for resource in resources:
				value = self.mip.value(self.x[(T,resource)])
				if value != None and value > 0:
					task_resources.append(resource)
			T.resources = task_resources

	def solve(self, scenario, bigm=10000, kind='CBC', time_limit=None, random_seed=None, ratio_gap=0.0, msg=0):

		self.scenario = scenario
		self.horizon = self.scenario.horizon
		self.bigm = bigm
		self.build_mip_from_scenario(msg=msg)

		params = dict()
		if time_limit is not None:
			params['time_limit'] = str(time_limit)
		if random_seed is not None:
			params['random_seed'] = str(random_seed)
		params['kind']= kind
		params['ratio_gap'] = str(ratio_gap)
		self.mip.solve(msg=msg,**params)
		#_solve_mip(self.mip, kind=kind, params=params, msg=msg)

		if self.mip.status() == 1:
			self.read_solution_from_mip(msg=msg)
			return 1
		if msg:
			print('ERROR: no solution found')
		return 0



class DiscreteMIP(object):
	"""
	pulp with time discretisation
	"""

	def __init__(self,mip):
		self.mip = mip
		self.scenario = None
		self.horizon = None
		self.task_groups = None
		self.x = None  # mip variables shortcut

	def build_mip_from_scenario(self, msg=0):
		S = self.scenario
		mip = self.mip
		self.task_groups = _get_task_groups(self.scenario)

		x = dict()  # mip variables
		cons = list()  # log of constraints for debugging
		for T in self.task_groups:
			task_group_size = len(self.task_groups[T])
			# base time-indexed variables
			cat = 'Binary'
			if task_group_size > 1:
				cat = 'Integer'
			x.update({ (T,t) : mip.var(str((T, t)), 0, task_group_size, cat) for t in range(self.horizon) })
			affine = [(x[T, t], 1) for t in range(self.horizon) ]

			# check if task is required
			if T.reward is None:
				cons.append(mip.con(affine, sense=0, rhs=task_group_size))
			else:
				cons.append(mip.con(affine, sense=-1, rhs=task_group_size))

			for RA in T.resources_req:
				# check if contains a single resource
				if len(RA) <= 1:
					continue
				# create variables if necessary except the first one
				x.update({ (T,R,t) : mip.var(str((T, R, t)), 0, task_group_size, cat)
						   for R in RA for t in range(self.horizon) if (T,R,t) not in x})
				# enough position needs to get selected
				affine = [(x[T, R, t], 1) for R in RA for t in range(self.horizon) ]
				# TODO: can the next line be removed?
				# cons.append(mip.con(affine, sense=0, rhs=task_group_size))
				# synchronize with base variables
				for t in range(self.horizon):
					affine = [(x[T, R, t], 1) for R in RA] + [(x[T,t],-1)]
					cons.append(mip.con(affine, sense=0, rhs=0))

			# generate shortcuts for single resources
			x.update({ (T,R,t) : x[T,t] for RA in T.resources_req if len(RA) == 1
					   for R in RA for t in range(self.horizon) })

		# synchronize in RA with multiple tasks
		ra_to_tasks = S.resources_req_tasks()
		for RA in ra_to_tasks:
			tasks = list(set(ra_to_tasks[RA]) & set(self.task_groups))
			if not tasks:
				continue
			for R in RA:
				T_ = tasks[0]
				for T in tasks[1:]:
					affine = [(x[T,R,t],1) for t in range(self.horizon) ]+\
							 [(x[T_,R,t],-1) for t in range(self.horizon) ]
					cons.append(mip.con(affine, sense=0, rhs=0))

		# everybody is finished before the horizon TODO: why is this check necessary
		affine = [ (x[T, t],1) for T in S.tasks() for t in range(self.horizon-T.length+1,self.horizon) if (T,t) in x ]
		cons.append(mip.con(affine, sense=-1, rhs=0))

		coeffs = { (T,R) : RA[R] for T in S.tasks() for RA in T.resources_req for R in RA }
		for R in S.resources():
			if R.size is not None:
				resource_size = R.size
			else:
				resource_size = 1.0
			for t in range(self.horizon):
				affine = [(x[T, R, t_], coeffs[T,R])
						  for T in set(S.tasks(resource=R)) & set(self.task_groups)
						  for t_ in range(t+1-T.length,t+1)
						  if (T,R,t_) in x ]
				cons.append(mip.con(affine, sense=-1, rhs=resource_size))

		# lax precedence constraints
		for P in S.precs_lax():
			if P.task_left not in self.task_groups or P.task_right not in self.task_groups:
				continue
			left_size = float(len(self.task_groups[P.task_left]))
			right_size = float(len(self.task_groups[P.task_right]))
			#in the default case it is expected that the task groups have similar
			#size, so the first task in the left task group must be scheduled before
			#the first task in the right task group, and so on
			if left_size == right_size:
				for t in range(max(-P.task_left.length-P.offset,0),min(self.horizon-P.task_left.length-P.offset,self.horizon)) :
					affine = \
						[(x[P.task_left, t_],1) for t_ in range(t,self.horizon)] + \
						[(x[P.task_right, t_],-1) for t_ in range(t+P.task_left.length+P.offset,self.horizon)]
					cons.append(mip.con(affine, sense=-1, rhs=0))
			#covers the case that the right task group has size ones
			#in which case this task should be scheduled after all tasks in the
			#left task group
			elif left_size == 1 or right_size == 1:
				for t in range(max(P.task_left.length+P.offset,0),min(self.horizon+P.task_left.length+P.offset,self.horizon)):
					affine = \
						[(x[P.task_left, t_],1/left_size) for t_ in range(t,self.horizon)] + \
						[(x[P.task_right, t_],-1/right_size) for t_ in range(t+P.task_left.length+P.offset,self.horizon)]
					cons.append(mip.con(affine, sense=-1, rhs=0))
			#covers the cases that the task groups have different sizes != 1
			#if the right task group is larger, then the  additonal tasks
			#will have no contraints
			else:
				for t in range(max(P.task_left.length+P.offset,0),min(self.horizon+P.task_left.length+P.offset,self.horizon)):
					affine = \
						[(x[P.task_left, t_],1) for t_ in range(t-P.task_left.length-P.offset)] + \
						[(x[P.task_right, t_],-1) for t_ in range(t)]
					cons.append(mip.con(affine, sense=1, rhs=0))

		# tight precedence constraints
		for P in S.precs_tight():
			if P.task_left not in self.task_groups or P.task_right not in self.task_groups:
				continue
			'''
			if P.offset >= 0:
				affine = [(x[P.task_left, t],t) for t in range(self.horizon)] + \
					 [(x[P.task_right, t],-t) for t in range(self.horizon)]
				pulpmip.cons(affine, sense=0, rhs=-(P.task_left.length + P.offset))
			elif P.offset < 0:
				affine = [(x[P.task_left, t],t) for t in range(self.horizon)] +\
					 [(x[P.task_right, t],-t) for t in range(self.horizon)]
				pulpmip.cons(affine, sense=0, rhs= P.task_right.length-P.offset )
			'''
			if P.offset >= 0:
				for t in range(self.horizon):
					affine = [(x[P.task_left, t_],1) for t_ in range(t,self.horizon)] + \
						 [(x[P.task_right, t_],-1) for t_ in range(min(t+P.task_left.length+P.offset,self.horizon),self.horizon)]
					cons.append(mip.con(affine, sense=0, rhs=0))
			elif P.offset < 0:
				for t in range(self.horizon):
					affine = [(x[P.task_right, t_],1) for t_ in range(t)] + \
						 [(x[P.task_left, t_],-1) for t_ in range(min(t+P.task_right.length-P.offset,self.horizon))]
					cons.append(mip.con(affine, sense=0, rhs=0))

		# low bounds
		for P in S.bounds_low():
			if P.task not in self.task_groups:
				continue
			affine = [(x[P.task, t],1) for t in range(P.bound)]
			cons.append(mip.con(affine, sense=0, rhs=0))

		# up bounds
		for P in S.bounds_up():
			if P.task not in self.task_groups:
				continue
			affine = [(x[P.task, t],1) for t in range(max(P.bound-P.task.length+1,0),self.horizon)]
			cons.append(mip.con(affine, sense=0, rhs=0))

		# tight low bounds
		for P in S.bounds_low_tight():
			if P.task not in self.task_groups:
				continue
			task_group_size = len(self.task_groups[P.task])
			cons.append(mip.con([(x[P.task, P.bound],1)], sense=0, rhs=task_group_size))

		# tight up bounds
		for P in S.bounds_up_tight():
			if P.task not in self.task_groups:
				continue
			task_group_size = len(self.task_groups[P.task])
			cons.append(mip.con([(x[P.task, max(P.bound-P.task.length,0)],1)], sense=0, rhs=task_group_size))

		# conditional precedence constraints
		for P in S.precs_cond():
			if P.task_left not in self.task_groups or P.task_right not in self.task_groups:
				continue
			shared_resources = list(set(S.resources(task=P.task_left)) & set(S.resources(task=P.task_right)))
			for R in shared_resources:
				for t in range(self.horizon):
					affine = [(x[P.task_left, R, t], 1)] +\
							 [(x[P.task_right, R, t_],1)
							  for t_ in range(t,min(t+P.task_left.length+P.offset,self.horizon))]
					rhs = (len(self.task_groups[P.task_left])+len(self.task_groups[P.task_right])/2.0)
					cons.append(mip.con(affine, sense=-1, rhs=rhs))

		# capacity lower bounds
		for C in S.capacity_low():
			# weight gets proportionally assigned according to overlap
			affine = [ (x[T, C.resource, t_], C.weight(T=T,t=t)/float(T.length) )
					  for t in range(self.horizon)
					  for T in self.task_groups
					  for t_ in range(max(0,t-T.length+1),t+1)
					  if (T,C.resource,t_) in x and C.weight(T=T,t=t) ]
			'''
			affine = [ (x[T, C.resource, t], C.weight(T=T,t=t)) for T in self.task_groups
					  for t in range(self.horizon) if (T,C.resource,t) in x and C.weight(T=T,t=t) ]
			'''
			if not affine:
				continue
			# sum up (pulp doesnt do this)
			affine_ = { a:0 for a,b in affine }
			for a,b in affine:
				affine_[a] += b
			affine = [ (a,affine_[a]) for a in affine_ ]
			cons.append(mip.con(affine, sense=1, rhs=C.bound))

		# capacity upper bounds
		for C in S.capacity_up():
			# weight gets proportionally assigned according to overlap
			affine = [ (x[T, C.resource, t_], C.weight(T=T,t=t)/float(T.length) )
					  for t in range(self.horizon)
					  for T in self.task_groups
					  for t_ in range(max(0,t-T.length+1),t+1)
					  if (T,C.resource,t_) in x and C.weight(T=T,t=t) ]
			'''
			affine = [ (x[T, C.resource, t], C.weight(T=T,t=t)) for T in self.task_groups
					  for t in range(self.horizon) if (T,C.resource,t) in x and C.weight(T=T,t=t) ]
			'''
			if not affine:
				continue
			# sum up (pulp doesnt do this)
			affine_ = { a:0 for a,b in affine }
			for a,b in affine:
				affine_[a] += b
			affine = [ (a,affine_[a]) for a in affine_ ]
			cons.append(mip.con(affine, sense=-1, rhs=C.bound))

		# capacity switch bounds
		count = 0
		for C in S.capacity_diff_up():
			R = C.resource
			# get affected periods
			periods = list(set([ t for t in range(self.horizon) for T in self.task_groups if C.weight(T,t) ]))
			periods = sorted(periods)
			x.update({ ('switch_%i'%count,R,t) : mip.var(str(('switch_%i'%count,R, t)), 0, 1)
					   for t in periods })
			# define switch variables
			for t in periods[:-1]:
				# decrease switch
				if C.kind == 'diff' or C.kind == 'diff_dec':
					affine = [ (x['switch_%i'%count,R,t],1) ] +\
							 [ (x[T,R,t-T.length+1],-1) for T in self.task_groups
							   if (T,R,t-T.length+1) in x and C.weight(T,t) ] +\
							 [ (x[T,R,t+1],1) for T in self.task_groups
							   if (T,R,t+1) in x and C.weight(T,t+1) ]
					cons.append(mip.con(affine, sense=1, rhs=0))
				# increase switch
				if C.kind == 'diff' or C.kind == 'diff_inc':
					affine = [ (x['switch_%i'%count,R,t],1) ] +\
							 [ (x[T,R,t-T.length+1],1) for T in self.task_groups
							   if (T,R,t-T.length+1) in x and C.weight(T,t) ] +\
							 [ (x[T,R,t+1],-1) for T in self.task_groups
							   if (T,R,t+1) in x and C.weight(T,t+1) ]
					cons.append(mip.con(affine, sense=1, rhs=0))
			affine = [ (x['switch_%i'%count,R,t],1) for t in periods[:-1] if ('switch_%i'%count,R,t) in x ]
			cons.append(mip.con(affine, sense=-1, rhs=C.bound))
			count += 1

		def task2cost(T,t):
			cost = 0
			if T.completion_time_cost is not None:
				cost += T.completion_time_cost*t
			if T.reward is not None:
				cost -= T.reward
			return cost

		objective = [
			(x[T, t], task2cost(T,t))
			for T in S.tasks()
			if T in self.task_groups
			for t in range(self.horizon) ]

		mip.obj(objective)
		'''
		for con in cons:
			mip.add_con(con)
		'''
		self.mip = mip
		self.x = x


	def read_solution_from_mip(self, msg=0):
		for T in self.task_groups:
			# get all possible starts with combined resources
			starts = \
				[ (t,R)
				for t in range(self.horizon)
				for R in self.scenario.resources()
				if (T,R,t) in self.x and self.mip.value(self.x[T, R, t]) is not None
				for i in range(int(self.mip.value(self.x[T, R, t])))
				]

			# iteratively assign starts and resources
			for T_ in self.task_groups[T]:
				# reset values
				T_.start_value = None
				T_.resources = list()
				# in case of not required tasks with reward, there might be less starts than tasks
				if not starts:
					break
				# consider single resources first
				RAs = [ RA for RA in T_.resources_req if len(RA) == 1 ] + \
					  [ RA for RA in T_.resources_req if len(RA) > 1  ]
				T_.start_value = [ t for (t, R) in starts ][0]
				for RA in RAs :
					if set(T_.resources) & set(RA):
						continue
					(t, R) = [(t_, R_) for (t_, R_) in starts if R_ in RA and t_ == T_.start_value][0]
					starts.remove((t, R))
					T_.resources.append(R)


	def solve(self, scenario, kind='CBC', time_limit=None, random_seed=None, ratio_gap=0.0, msg=0):

		self.scenario = scenario
		if self.scenario.horizon is None:
			raise Exception('ERROR: solver requires scenarios with defined horizon')
			return 0
		self.horizon = self.scenario.horizon
		self.build_mip_from_scenario(msg=msg)

		# if time_limit :
		#   options += ['sec',str(time_limit),'ratioGap',str(0.1),'cuts','off',
				#       'heur','on','preprocess','on','feas','on']#,'maxNodes',str(0),'feas','both','doh','solve']

		params = dict()
		if time_limit is not None:
			params['time_limit'] = time_limit
		if random_seed is not None:
			params['random_seed'] = str(random_seed)
		#params['cuts'] = 'off'
		params['ratio_gap'] = str(ratio_gap)
		params['kind'] = kind
		self.mip.solve(msg=msg,**params)

		#_solve_mip(self.mip, kind=kind, params=params, msg=msg)

		if self.mip.status() == 1:
			self.read_solution_from_mip(msg=msg)
			return 1
		if msg:
			print('ERROR: no solution found')
		return 0
